import { defineStore } from 'pinia';
import { ref } from 'vue';
import { useGetServer } from "@/composables/useGetServer.ts";

export const usePartyLineStore = defineStore('partyLine', () => {
    const partyLine = ref('');
    const rumor = ref('');
    const eventSource = ref<EventSource | null>(null);
    const partyLines = ref<string[]>([]);
    const partyLineDeletedFlag = ref(false);

    const connectEventSource = async (currentPartyLine: string) => {
        if (eventSource.value && partyLine.value === currentPartyLine) {
            console.log(`Already connected to the party line: ${currentPartyLine}`);
            return;
        }
        await disconnectCurrentPartyLine(); // Ensure we disconnect first

        console.log(`Connecting to party line: ${currentPartyLine}`);
        eventSource.value = new EventSource(`${useGetServer()}/connectPartyLine?partyLine=${encodeURIComponent(currentPartyLine)}`);

        eventSource.value.addEventListener('open', () => {
            console.log('Connected to party line:', currentPartyLine);
        });

        eventSource.value.addEventListener('message', (event) => {
            console.log('Rumor received:', event.data);
            rumor.value = event.data;
        });

        eventSource.value.addEventListener('error', (err) => {
            console.error('EventSource error:', err);
            if (rumor.value.includes('PARTY_LINE_DELETED')) {
                disconnectCurrentPartyLine();
                partyLineDeletedFlag.value = true;
            } else {
                eventSource.value!.close();
                setTimeout(() => connectEventSource(currentPartyLine), 5000); // Attempt reconnection after 5 seconds
            }
        });
    };

    const disconnectCurrentPartyLine = async () => {
        if (eventSource.value) {
            console.log(`Disconnecting from party line: ${partyLine.value}`);
            eventSource.value.close();
            eventSource.value = null;
            partyLine.value = '';
            rumor.value = '';
            console.log('Disconnected successfully');
        } else {
            console.log('No active connection to disconnect');
        }
    };

    const fetchPartyLines = async () => {
        try {
            const response = await fetch(useGetServer() + '/partyLines');
            const data = await response.json();
            partyLines.value = data.allPartyLines;
        } catch (error: any) {
            console.error('Error fetching party lines:', error);
        }
    };

    const createPartyLine = async (partyLine: string) => {
        try {
            const response = await fetch(useGetServer() + '/createPartyLine', {
                method: 'POST',
                headers: {
                    'Content-Type': 'application/json',
                },
                body: JSON.stringify({ partyLine }),
            });

            const data = await response.json();

            if (response.ok) {
                await fetchPartyLines(); // Fetch the updated list of party lines
            } else {
                console.error('Failed to create party line');
            }

            return data;
        } catch (error: any) {
            console.error('Error creating party line:', error);
        }
    };

    const deletePartyLine = async (partyLine: string) => {
        try {
            const response = await fetch(useGetServer() + '/deletePartyLine', {
                method: 'DELETE',
                headers: {
                    'Content-Type': 'application/json',
                },
                body: JSON.stringify({ partyLine }),
            });

            if (response.ok) {
                console.log(`Deleted party line: ${partyLine}`);
                await fetchPartyLines(); // Refresh the party lines after deletion
            } else {
                console.error('Failed to delete party line');
            }
        } catch (error: any) {
            console.error('Error deleting party line:', error);
        }
    };

    return {
        rumor,
        partyLine,
        partyLines,
        eventSource,
        partyLineDeletedFlag,
        connectEventSource,
        disconnectCurrentPartyLine,
        fetchPartyLines,
        createPartyLine,
        deletePartyLine,
    };
});
